
#include "avr/interrupt.h"
#include "avr/sleep.h"
#include "avr/wdt.h"

// possible increase from 30 to 50 for 40% increase in current on POTRB pin 4
#define IR_SEND_DUTY_CYCLE 30
#include <IRremote.h>
IRsend IrSender;

/*  Watchdog Timer and sleep relationship
 *  The sleep increment can be values between 16ms and 8s.  All increments are
 *  powers of two starting at 16.  So: 16, 32, 64, 128, 256, 512, ... 8192 milliseconds
 *  are possible.  Use Equation 16 * 2 ^ (WATCHDOG_TIMER_PRESCALER - 1)
 *  Eg: for a sleep increment of about 250ms use a WATCHDOG_TIMER_PRESCALER of 4
 */
#ifdef HALF_RESOLUTION
// see other HALF_RESOLUTION code options for explination
#define WATCHDOG_TIMER_PRESCALER 5
#else
#define WATCHDOG_TIMER_PRESCALER 4
#endif
// using a 250ms watchdog timer gives 30 samples for about seven seconds
#define MAX_SAMPLE 30
uint8_t samples[MAX_SAMPLE];
uint8_t ptr = 0;

/*  Watchdog Timeout Interrupt Enable
 *  normally the watchdog timeout will reset the MCU, this is not
 *  desirable and is the persistent action the hardware will take
 *  when the Watchdog Timer is enabled.
 *  
 *  For the *next* reset we can instead interrupt the MCU using the
 *  Watchdog Timeout Interrupt Enable bit.  This bit needs to be set
 *  at least once per sleep cycle
 */
#define INTERRUPT_NEXT_RESET() { WDTCR |= (1UL << WDIE); }

#define PORTB_ZERO_ON() { PORTB |= (1UL << PB0); }
#define PORTB_ZERO_OFF() { PORTB &= ~(1UL << PB0); }


#define FLICKER_WHEN_USING_POWER

#define FLICKER_THRESHOLD 15
#define MINIMUM_SLEEP_MS 2500
#define MINIMUM_POWER_MS 6000

enum { SLEEPING, AWAKE };
bool state = AWAKE;

enum { USING_POWER = 0, NOT_USING_POWER, NONE };
volatile uint8_t change = NONE;
volatile uint8_t wtdTimer = 1;

uint32_t lastEvent;
uint16_t total;
uint8_t last = USING_POWER;
uint32_t changes = 0;

/*  When using a weak detector, it may take a long time to measure power,
 *  in this case it makes sense to clear all previously recorded sleep
 *  since, in reality, the monitored device was in fact awake.
 *  
 *  When using a sensitive detector, this can make it difficult to detect
 *  when the device has gone to sleep.  Turn off the default behavior of 
 *  clearing previously recorded sleep by defining NO_HAIR_TRIGGER_WHEN_AWAKE
 *  
 *  If the sensitive detector is very sensitive, consider using LAZY_WHEN_SLEEPING
 *  to prevent waking up.
 */
#define NO_HAIR_TRIGGER_WHEN_AWAKE
//#define LAZY_WHEN_SLEEPING

/*  convert output pin to enumerated USING_POWER or NOT_USING_POWER
 *  the opto-isolator inverts the signal so input is low when USING_POWER
 *  and high when NOT_USING_POWER
 *  
 *  [!] there is no need to invert the logic here since enum has USING_POWER
 *      equal to zero (invert would be 0 == (PINB & (1UL << PINB3)) >> PINB3)
 */
__attribute__((always_inline)) static inline
uint8_t readInput()  { return (PINB & (1UL << PINB3)) >> PINB3; }

void setup() {

  resetSamples();
  //resetDebounce(-1);
  lastEvent = millis();
  
  pinMode(0, OUTPUT); PORTB_ZERO_ON();
  pinMode(1, /* not used */ INPUT_PULLUP);
  pinMode(2, /* not used */ INPUT_PULLUP);
  pinMode(3, INPUT); // DDRB &= ~(1 << PB3);
  pinMode(4, OUTPUT); /* used by IRsend */

  /*  Turn off the ADC and set the sleep mode that will be used
   *  later on.  Use sleep_mode() macro to actually go to sleep
   */
  ADCSRA &= ~(1UL << ADEN);
  // [!] do not use, cannot tell how long device has been sleeping
  //set_sleep_mode(SLEEP_MODE_PWR_DOWN);
  set_sleep_mode(SLEEP_MODE_IDLE);   

  /* Interrupt setup for a 250ms event - uses watchdog timer
   * 
   */
  { // clear the Watchdog Reset Flag
    MCUSR &= ~(1UL << WDRF);
    // configure the Watchdog Timer Prescaler (clear all other bits)
    WDTCR = WATCHDOG_TIMER_PRESCALER;
    
    // enable the Watchdog Timer
    WDTCR |= (
      (1UL << WDCE) | /* Watchdog Change Enable */
      (1UL << WDE)    /* Watchdog Enable */
    );
  }

  /*  Interrupt setup for the ATTINY85 MCU - portb pin 4
   */
  // enable the pin change interrupt
  GIMSK |= (1UL << PCIE);
  // interrupt on pin change for PB3
  PCMSK |= (1UL << PCINT3);
  sei();

}

// record the value of PORTB pin 3
ISR(PCINT0_vect) {
  /*  change is enumerated:
   *  0 - USING_POWER
   *  1 - NOT_USING_POWER
   *  2 - NONE (external use only)
   *  
   *  [!] power indicator inverted, NPN in photo-transistor inverts signal
   *  when PINB pin 3 is high, monitored device uses no measurable power
   *  when low, monitored deviec is using power
   */
  change = readInput();
}

ISR(WDT_vect) {
  wtdTimer = 0;
}

void toggleLgDisplay() {
  unsigned long data = 0x20F0B54A;
    // send power three times, NEC IR command does not have separate on/off
    for(uint8_t i = 3; i--;) IrSender.sendNEC(data, 32);
}

/*  halt execution of the program for a little while
 *  the Amazon Fire has some incompatibilities with the first version of the hardware
 *  due to weak power usage detection as well as weird power usage when sleeping / waking
 *  * waking from sleep: uses very little power, I assume most applications are paused
 *    so we cannot really detect that power is being used
 *  * sleeping: sometimes wakes up for a brief period without sending anything to the display
 *    this is different from the pulse of power it uses when sleeping
 *  halt for 30 seconds
 */
void haltProgram() {
  // using the quarter-second watchdog timer, halt for between 0 and 63 seconds
  for(uint8_t i = 120 /* divide by four = seconds */; i--;) {
    INTERRUPT_NEXT_RESET();
    // the Arduino millis interrupt about once per millisecond, go back to sleep
    while(wtdTimer) sleep_mode();
    wtdTimer = 1;
  }
}

void resetSamples() {
  for(uint8_t i = MAX_SAMPLE; i--;) samples[i] = 0;
  total = 0;
}

void loop() {
  // change next watchdog reset into an interrupt
  INTERRUPT_NEXT_RESET();
  /*  put the mcu into low power mode, come out a few times a second for processing
   *  
   *  [!] the Arduino millis() function configures timer0 to interrupt about once per
   *  millisecond causing the SLEEP_MODE_IDLE to not really be very useful.  And the IR
   *  library needs millis() to work for some reason.  Also I think we are going to use
   *  it too, so we need to monitor these unwanted wake-up events.  So keep putting MCU
   *  sleep on all but pin change and watchdog timer interrupts
   */
  while(wtdTimer && (NONE == change)) sleep_mode();

  // immediately make a copy of change, and reset the interrupt maintained varaible
  uint8_t level = change;
  change = NONE;

  // when there is a change, set store the change in memory to prevent needing to read PINB
  if(NONE != level) last = level;

  // the delay may not have completed successfully due to an interrupt
  uint32_t now = millis();
  uint32_t temp, delta = now - lastEvent;
  lastEvent = now;

#ifdef HALF_RESOLUTION
  /*  originally te watchdog timer caused an interrupt every quarter-second
   *  in reduced resolution mode time is slowed down by 1/2 so that we can
   *  still meet memory requirements for ATTINY85 and store enough time to
   *  track the Amazon Fire using the prototype power tracking hardware
   */
  // round up (this needs to happen before compression)
  delta += (delta & 1);
  // reduce delta so it can fit into a single byte (compression)
  delta /= 2;
#endif
  
  #ifdef NO_HAIR_TRIGGER_WHEN_AWAKE
  #else
  // when awake any change will clear out all previous sleep tracking
  if((AWAKE == state) && (NONE != level)) resetSamples();
  #endif
  
  // handle any sub quarter-second interrupts (otherwise reset wtdTimer)
  if(0 == wtdTimer) {
    // reset the watchdog timer flag
    wtdTimer = 1;
  } else {
    // flicker filter tracking
    changes++;
    /*  while sleeping - only track power usage, when awake - the opposite
     *   
     *  [?] this means that if we are awake, and a change has put us into NOT_USING_POWER
     *      we have to wait until USING_POWER to be able to record how long we were
     *      NOT_USING_POWER - kinda confusing
     */
    temp  = (SLEEPING == state) && (NOT_USING_POWER == level);
    temp |= (AWAKE == state)    && (USING_POWER == level);
    if(temp) {
      // pre-calculate the value of the integrated sample as to not let it exceed 8-bits
      temp = delta + samples[ptr];
      samples[ptr] = (temp > 250) ?250 :temp;
    }
    // do not proceed on sub quarter-second interrupts (go back to sleep)
    return;
  }

  // on whole quarter-seconds we need the current level, NONE is not helpful
  uint8_t current = last;
  /*   on whole quarter-seconds we need the current level, NONE is not helpful
   *   
   *   [!] in the off chance we are chasing a change, the current level is incorrect, but
   *  can easily be fixed by inverting
   */
  if(NONE != level) current = !current;

  /*  flicker filter
   *  under some circumstances the input can change very rapidly.  In a quarter second it is possible
   *  for an 8MHz MCU to react to over 2 million pin changes.
   *  
   *  The integration software was not really designed for this type of tracking, even at the highest
   *  resolution changes were only really expected to happen once per millisecond limiting the
   *  software to only work in situations where a pin change only happens a thousand times per second.
   *  
   *  during testing I found that the digital Schmitt triggered hall effect sensor could switch 
   *  well over a hundred times, in this case I think the output should be pulled to either zero or 250
   *  
   *  [?] high fluctuation can be interperated two ways, in the case that fluctuation occur
   *      when USING_POWER:
   *        when high fluctuation in input and
   *          AWAKE - assume that we were that we were awake the whole time
   *          ASLEEP - assme that we were awake the whole time
   *      when NOT_USING_POWER
   *          AWAKE - assume that we were that we were sleeping the whole time
   *          ASLEEP - assme that we were sleeping the whole time
   */
  if(FLICKER_THRESHOLD < changes) {
    #ifdef FLICKER_WHEN_USING_POWER
    if(AWAKE == state) samples[ptr] = 0;
    else /* SLEEPING */ {
      // a higer flicker threshold should be used to go from SLEEPING to AWAKE
      if((2 * FLICKER_THRESHOLD) < changes) samples[ptr] = 250;
      else samples[ptr] = 0;
    }
    #else
    // do whatever is needed
    #endif
  }
  changes = 0;

  // integrate monitored device's power usage
  { uint8_t head = 1 + ptr;
    // make sure head wraps to zero when it reaches the end of the storage
    if(MAX_SAMPLE <= head) head = 0;
    // remove the last recorded value from the sample (currently pointed to by head)
    if(total > samples[head]) total -= samples[head]; else total = 0;
    // clear the value from storage
    samples[head] = 0;
    /*  while sleeping - only track power usage, when awake - the opposite
     *  
     *  [?] in this case, we know that since the last change we are *still* either
     *      USING_POWER or NOT_USING_POWER.  So tracking is straightforward.
     */
    temp  = (SLEEPING == state) && (USING_POWER == current);
    temp |= (AWAKE == state)    && (NOT_USING_POWER == current);
    if(temp) {
      // pre-calculate the value of the integrated sample as to not let it exceed 8-bits
      temp = delta + samples[ptr];
      samples[ptr] = (temp > 250) ?250 :temp;
    };
    total += samples[ptr];
    ptr = head;
  }

  if(AWAKE == state) {
    #ifdef NO_HAIR_TRIGGER_WHEN_AWAKE
    #else
    // whenever AWAKE and USING_POWER it should be impossible to go to sleep
    if(USING_POWER == current) resetSamples();
    #endif
    /*  when the monitored device is powered on there needs to be a very
     *  long period of non-measurable power consumption
     */
    if(total >= MINIMUM_SLEEP_MS) {
      state = SLEEPING; PORTB_ZERO_OFF(); resetSamples();
      // propoagte sleep signal to other devices: {...}
      {
        // send signal to the LG LED Projector
        toggleLgDisplay();
      }
    }
  } else /* SLEEPING */ {
    #ifdef LAZY_WHEN_SLEEPING
    // whenever SLEEPING and NOT_USING_POWER it should be difficult to wake up
    if(NOT_USING_POWER == current) resetSamples();
    #else
    #endif
    /*  when the monitored device is sleeping or powered off only a small amount
     *  power consumption is needed to switch to awake
     */
    if(total >= MINIMUM_POWER_MS) {
      state = AWAKE; PORTB_ZERO_ON(); resetSamples();
      // propoagte wake signal to other devices: {...}
      {
        // send signal to the LG LED Projector
        toggleLgDisplay();
      }
    }
  }

}
